/* H5400 Fingerprint Sensor Interface client library
 * Copyright (c) 2004 Jrgen Andreas Michaelsen <jorgenam@ifi.uio.no> 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <string.h>
#include <stdlib.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include "fsiclient.h"

struct translation_element
{
  int dx;
  int dy;
  unsigned int m;
};

static void compute_motion( unsigned char *frame_ref, 
			    unsigned char *frame_new,
			    struct translation_element *te )
{
  te->m = (__typeof__(te->m))-1;
  te->dx = 0;
  te->dy = 0;

  for( int dx = -2; dx <= 2; dx++ )
    for( int dy = 0; dy < 4; dy++ )
      {
	unsigned int sad = 0;

	for( int y = 0; y <= 4; y++ )
	  for( int x = 12; x < 268; x++ )
	    sad += abs( frame_new[y*FSI_FRAME_WIDTH+x] - 
			frame_ref[x+dx+(FSI_FRAME_WIDTH*(y+dy))] );

	if( sad < te->m )
	  {
	    te->m = sad;
	    te->dx = dx;
	    te->dy = dy;
	  }
      }
}

static int make_raw_frame( unsigned long *src, unsigned char *dst )
{
  unsigned int temperature;
  const long sync_col = src[FSI_FRAME_SIZE-1];

  if( FSI_DB1_EVEN(sync_col) != 0x0 || (FSI_DB1_ODD(sync_col)&0xE) != 0xE ||
      FSI_DB2_EVEN(sync_col) != 0x0 || (FSI_DB2_ODD(sync_col)&0xE) != 0xE )
    return -1;

  temperature = FSI_DB3_EVEN( sync_col );

  for( int x = 0; x < FSI_FRAME_WIDTH; x++ )
    {
      dst[x+0*FSI_FRAME_WIDTH] = FSI_DB4_EVEN( src[x] );
      dst[x+1*FSI_FRAME_WIDTH] = FSI_DB4_ODD ( src[x] );
      dst[x+2*FSI_FRAME_WIDTH] = FSI_DB3_EVEN( src[x] );
      dst[x+3*FSI_FRAME_WIDTH] = FSI_DB3_ODD ( src[x] );
      dst[x+4*FSI_FRAME_WIDTH] = FSI_DB2_EVEN( src[x] );
      dst[x+5*FSI_FRAME_WIDTH] = FSI_DB2_ODD ( src[x] );
      dst[x+6*FSI_FRAME_WIDTH] = FSI_DB1_EVEN( src[x] );
      dst[x+7*FSI_FRAME_WIDTH] = FSI_DB1_ODD ( src[x] );
    }

  return temperature;
}

static void add_frame( unsigned char *output, 
		       unsigned char *frame, 
		       unsigned int *contrast, 
		       struct translation_element *te, 
		       struct translation_element *gte )
{
  int start = 0;
  int stop  = FSI_FRAME_WIDTH;

  gte->dx += te->dx;

  if( gte->dx < 0 )
    start += -gte->dx;
  else if( gte->dx > 0 )
    stop -= gte->dx;

  for( int y = 0; y < FSI_FRAME_HEIGHT; y++ ) 
    {
      for( int x = 0; x < FSI_FRAME_WIDTH; x++ )
	output[y*FSI_FRAME_WIDTH+x] += 
	  frame[x+(x >= start && x < stop ? gte->dx : -gte->dx) +
		FSI_FRAME_WIDTH*y];

      contrast[y] += 0xf;
    }
}

static int compute_translations( unsigned long *input, 
				 unsigned int frame_count,
				 struct translation_element *te )
{
  unsigned char tmp_frame_buf[2][FSI_FRAME_WIDTH*FSI_FRAME_HEIGHT];
  unsigned int this_frame = 0;
  unsigned int line_count = FSI_FRAME_HEIGHT;
  unsigned int temperature;

  for( unsigned int current_frame = 0; 
       current_frame < frame_count; current_frame++ )
    {
      int frame_temp;
      frame_temp = make_raw_frame( input, tmp_frame_buf[this_frame] );
      if( frame_temp < 0 )
	return -1;

      temperature += frame_temp;

      if( current_frame > 0 )
	{
	  compute_motion( tmp_frame_buf[!this_frame], 
			  tmp_frame_buf[this_frame], &te[current_frame] );

	  line_count += te[current_frame].dy;
	}
      else
	{
	  te[0].dx = 0;
	  te[0].dy = 0;
	}

      this_frame = !this_frame;
      input += FSI_FRAME_SIZE;
    }

  te[0].m = temperature;
  return line_count;
}

static void assemble_frames( unsigned long *input,
			     unsigned char *output,
			     unsigned int frame_count,
			     unsigned int line_count,
			     struct translation_element *te )
{
  unsigned char tmp_frame_buf[FSI_FRAME_WIDTH*FSI_FRAME_HEIGHT];
  unsigned char tmp_output_buf[FSI_FRAME_WIDTH*line_count];
  unsigned char *output_ptr = tmp_output_buf;
  unsigned int contrast[line_count];
  unsigned int cpos = 0;

  memset( tmp_output_buf, 0, sizeof(tmp_output_buf) );
  memset( contrast, 0, sizeof(contrast) );

  for( unsigned int current_frame = 0; 
       current_frame < frame_count; current_frame++, input += FSI_FRAME_SIZE )
    {
      output_ptr += te[current_frame].dy * FSI_FRAME_WIDTH;
      cpos += te[current_frame].dy;

      make_raw_frame( input, tmp_frame_buf );
      add_frame( output_ptr, tmp_frame_buf, &contrast[cpos], 
		 &te[current_frame], &te[0] );
    }

  for( unsigned int y = 0; y < line_count; y++ )
    for( int x = 0; x < FSI_FRAME_WIDTH; x++ )
      output[y*FSI_FRAME_WIDTH+x] = tmp_output_buf[y*FSI_FRAME_WIDTH+x] * 
	0xFF / contrast[y];
}

char *fsi_get_error_string( struct fsi_context *ctx )
{
  int tmperrno;

  strncat( ctx->err.string, ": ", sizeof ctx->err.string );

  switch( ctx->err.code & 0xFFFF00 )
    {
    case FSI_ERROR_OPEN:
      strncat( ctx->err.string, "open", sizeof ctx->err.string );
      break;
    case FSI_ERROR_READ:
      strncat( ctx->err.string, "read", sizeof ctx->err.string );
      break;
    case FSI_ERROR_CLOSE:
      strncat( ctx->err.string, "close", sizeof ctx->err.string );
      break;
    case FSI_ERROR_IOCTL:
      strncat( ctx->err.string, "ioctl", sizeof ctx->err.string );
      break;
    case FSI_ERROR_SYNC:
      strncat( ctx->err.string, "sync error", sizeof ctx->err.string );
      break;
    }

  tmperrno = ctx->err.code & 0xFF;

  if( tmperrno )
    {
      strncat( ctx->err.string, ": ", sizeof ctx->err.string );
      strncat( ctx->err.string, strerror(tmperrno), sizeof ctx->err.string );
    }

  return ctx->err.string;
}

int fsi_init( struct fsi_context *ctx )
{
  memset( ctx, 0, sizeof *ctx );
  ctx->line_width = FSI_FRAME_WIDTH;
  strncpy( ctx->err.string, __func__, sizeof ctx->err.string );

  ctx->fd = open( FSI_DEV_NAME, O_RDONLY );
  if( ctx->fd < 0 )
    {
      ctx->err.code = FSI_ERROR_OPEN | errno;
      return -1;
    }

  return 0;
}

int fsi_read_frames( struct fsi_context *ctx, 
		     unsigned long *frames_buffer, 
		     unsigned int buffer_size )
{
  ssize_t rval;

  strncpy( ctx->err.string, __func__, sizeof ctx->err.string );

  ctx->frames_buffer = frames_buffer;
  rval = read( ctx->fd, frames_buffer, buffer_size );
  if( rval < FSI_FRAME_SIZE_BYTES ) /* if not enough data */
    {
      ctx->err.code = FSI_ERROR_READ | errno;
      return -1;
    }

  ctx->frame_count = rval / FSI_FRAME_SIZE_BYTES;

  return 0;
}

int fsi_merge_frames( struct fsi_context *ctx )
{
  struct translation_element te[ctx->frame_count];
  int line_count;

  line_count = compute_translations( ctx->frames_buffer, 
				     ctx->frame_count, te );
  if( line_count < 0 )
  {
    strncpy( ctx->err.string, __func__, sizeof ctx->err.string );
    ctx->err.code = FSI_ERROR_SYNC;
    return -1;
  }
  ctx->line_count = line_count;
  ctx->temperature = te[0].m;

  assemble_frames( ctx->frames_buffer, (unsigned char *)ctx->frames_buffer,
		   ctx->frame_count, line_count, te );

  return 0;
}

int fsi_release( struct fsi_context *ctx )
{
  int result;

  result = close( ctx->fd );
  if( result < 0 )
    {
      strncpy( ctx->err.string, __func__, sizeof ctx->err.string );
      ctx->err.code = FSI_ERROR_CLOSE | errno;
      return -1;
    }

  return 0;
}

unsigned int fsi_get_line_count( struct fsi_context *ctx )
{
  return ctx->line_count;
}

unsigned int fsi_get_line_width( struct fsi_context *ctx )
{
  return ctx->line_width;
}

int fsi_get_temperature( struct fsi_context *ctx )
{
  return ((ctx->temperature/ctx->frame_count)-8) * 1400;
}

unsigned int fsi_get_quality( struct fsi_context *ctx )
{
  return ctx->quality;
}

