/* H5400 Fingerprint Sensor Interface client library usage example
 * Copyright (c) 2004 Jrgen Andreas Michaelsen <jorgenam@ifi.uio.no> 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>

#include "fsiclient.h"

#define BUF_SIZE  (FSI_FRAME_SIZE_BYTES * 512)


void write_pgm_file( unsigned char *buffer, int width, int height, FILE *fp )
{
  int x, y;

  fprintf( fp, "P5\n" );
  fprintf( fp, "%d %d\n", width, height );
  fprintf( fp, "255\n" );

  for( y = height; y > 0; y-- )
    for( x = width; x > 0; x-- )
      fprintf( fp, "%c", buffer[(x-1) + (y-1) * width] );
}


int main( int argc, char **argv )
{
  int result;
  void *buffer;
  unsigned int num_lines;
  unsigned int line_width;
  char fname[256];
  FILE *fp;
  struct fsi_context ctx;
  struct timeval tv_start;
  struct timeval tv_now;


  if( argc == 2 )
    strncpy( fname, argv[1], sizeof(fname) );
  else
    sprintf( fname, "fp_%08lx.pgm", time(NULL) );

  /* allocate buffer for fingerprint (any "large" buffer will do) */

  buffer = malloc( BUF_SIZE );
  if( buffer == NULL )
    {
      fprintf( stderr, "malloc failed\n" );
      return 9;
    }

  /* obtain fingerprint image using library */

  result = fsi_init( &ctx );
  if( result < 0 )
    {
      fprintf( stderr, "%s\n", fsi_get_error_string(&ctx) );
      return 1;
    }

  printf( "swipe finger... " ); fflush( stdout );

  result = fsi_read_frames( &ctx, buffer, BUF_SIZE );
  if( result < 0 )
    {
      fprintf( stderr, "%s\n", fsi_get_error_string(&ctx) );
      return 2;
    }

  printf( "OK [dT=%d mK]\n", fsi_get_temperature(&ctx) );
  printf( "creating fingerprint image... " ); fflush( stdout );

  gettimeofday( &tv_start, NULL );
  result = fsi_merge_frames( &ctx );
  gettimeofday( &tv_now, NULL );
  if( result < 0 )
    {
      fprintf( stderr, "%s\n", fsi_get_error_string(&ctx) );
      return 3;
    }

  num_lines = fsi_get_line_count( &ctx );
  if( num_lines == 0 )
    return 0;

  line_width = fsi_get_line_width( &ctx );

  printf( "OK [size=%dx%d,Q=%d,dt=%ld]\n", line_width, num_lines, 
	  fsi_get_quality(&ctx), 
	  (tv_now.tv_sec-tv_start.tv_sec)*1000000 +
	  tv_now.tv_usec-tv_start.tv_usec );

  result = fsi_release( &ctx );
  if( result < 0 )
    {
      fprintf( stderr, "%s\n", fsi_get_error_string(&ctx) );
      return 5;
    }

  /* write image to file (pgm format) */

  buffer = realloc( buffer, num_lines * line_width ); /* trim buffer */
  if( buffer == NULL )
    {
      /* is this possible? (request is smaller than original size) */
      fprintf( stderr, "realloc failed\n" );
      return 8;
    }

  printf( "writing image to file... " ); fflush( stdout );

  fp = fopen( fname, "wb" );
  if( fp == NULL )
    return 4;

  write_pgm_file( buffer, line_width, num_lines, fp );

  printf( "OK [name=%s]\n", fname );
  fclose( fp );

  /* destroy buffer */

  free( buffer );
}
